library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;


entity power is
	generic
	(
	   -- Input and output size is 8 bit
		size : positive := 8
	);
	port
	(
		clock    					: in 	std_logic;
		base							: in 	std_logic_vector(size-1 downto 0);
		exponent                : in 	std_logic_vector(size-1 downto 0);
		result                  : out std_logic_vector(size-1 downto 0);
		enable                  : in  std_logic;
		ready                   : out std_logic;
		valid                   : out std_logic
	);		
end power;

-- The square & multiply algorithmen as a state machine
-- for calculation of a^b mod 2 ** m
architecture power_rtl of power is
	-- The states for the state machine
	type states is (zero, first, second, third, forth);
	-- The state machine starts at the left state
	signal state 		: states := zero;
	subtype long is natural range 2 ** size - 1 downto 0;

	signal res, exp 				: long;
	signal res_temp, exp_temp 	: long;
	subtype index is natural range size-1 downto 0;
	signal i 						: index;
begin	
	

	square_and_multiply : process (clock)
	begin
	   if rising_edge(clock)
		then   
			case state is
				-- The output is not valid, but the machine is ready
			   when zero  =>  valid <= '0';
									ready  <= '1';
									if enable = '1'
									then
										state <= first;
									end if;
				-- The machine is enabled and running
				when first =>	i   <= exponent'right;
									res <= 1;
									exp <= to_integer(unsigned(base));
									state <= second;
				-- The machine is calculating inside a loop
				when second => if i <= exponent'left
									then
										if exponent(i) = '1'
										then
											res_temp <= (res * exp) mod 2**size;										
										else
											res_temp <= res;
										end if;	
										exp_temp <= (exp * exp) mod 2**size;										
									end if;
									state <= third;
				-- Loop again or end the calculation ?
				when third =>	if i < exponent'left
									then
										res <= res_temp;
										exp <= exp_temp;										
									   i <= i + 1;
										state <= second;
									else
										state <= forth;
									end if;
				-- The result is valid and the machine goes back to wait state
				when forth => 	result <= std_logic_vector(to_unsigned(res,size));
									ready <= '0';
									valid <= '1';
									state <= zero;
			end case;
		end if; -- rising_edge
	end process square_and_multiply;

end architecture power_rtl;